import cv2
import matplotlib.pyplot as plt
import numpy as np
import math
import glob
import os
from PIL import Image  
import PIL 
import copy
import tensorflow as tf
from keras.models import load_model
import copy
from keras.preprocessing import image
 
# Load model
# loaded_model = tf.keras.models.load_model('20191126_full_model.h5')


def extract_oysters_from_image(image_path, plot=False):
    
    img = plt.imread(image_path)
    img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    
    # bilateral filter, edge-preserving
    bilateral_filtered = cv2.bilateralFilter(img_gray, 9,150,150)
    
    # threshold
    _, thresh = cv2.threshold(bilateral_filtered, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)
    
    # edge detection
    edges = cv2.Canny(thresh, 40, 70)
    
    # gaussian blur
    gaussian_blurred = cv2.GaussianBlur(edges,(5,5),cv2.BORDER_DEFAULT)
    
    # dilate to close contours
    dilated = cv2.dilate(gaussian_blurred,np.ones((3,3),np.uint8))
    
    contours, _ = cv2.findContours(dilated, cv2.RETR_EXTERNAL,
                              cv2.CHAIN_APPROX_SIMPLE)
    contours = sorted(contours, key=cv2.contourArea, reverse=True)
    
    bounding_rectangles = [cv2.boundingRect(c) for c in contours]
    
    if plot:
        plt.figure(figsize=(20,10))
        plt.imshow(img)
        plt.title("original")
        
        plt.figure(figsize=(20,10))
        plt.imshow(bilateral_filtered,cmap="gray")
        plt.title("bilateral filtered, grayscaled")
        
        plt.figure(figsize=(20,10))
        plt.imshow(thresh,cmap="gray")
        plt.title("threshold")
        
        plt.figure(figsize=(20,10))
        plt.imshow(edges,cmap="gray")
        plt.title("edges")
        
        plt.figure(figsize=(20,10))
        plt.imshow(gaussian_blurred,cmap="gray")
        plt.title("gaussian blurred")
        
        plt.figure(figsize=(20,10))
        plt.imshow(dilated,cmap="gray")
        plt.title("dilated")
        
        plt.figure(figsize=(20,10))
        img_gray = cv2.drawContours(img_gray, contours, -1, (0,255,0), 12)
        plt.imshow(img_gray,cmap="gray")
        plt.title("Contours")
    
    return bounding_rectangles
    
def step_by_step_images(image_path, plot=False):
    
    img = plt.imread(image_path)
    img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    
    # bilateral filter, edge-preserving
    bilateral_filtered = cv2.bilateralFilter(img_gray, 9,150,150)
    
    # threshold
    _, thresh = cv2.threshold(bilateral_filtered, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)
    
    # edge detection
    edges = cv2.Canny(thresh, 40, 70)
    edges_out = copy.deepcopy(edges)
    # gaussian blur
    gaussian_blurred = cv2.GaussianBlur(edges,(5,5),cv2.BORDER_DEFAULT)
    
    # dilate to close contours
    dilated = cv2.dilate(gaussian_blurred,np.ones((3,3),np.uint8))
    
    contours, _ = cv2.findContours(dilated, cv2.RETR_EXTERNAL,
                              cv2.CHAIN_APPROX_SIMPLE)
    contours = sorted(contours, key=cv2.contourArea, reverse=True)
    
    bounding_rectangles = [cv2.boundingRect(c) for c in contours]
    
    # paint bounding boxes
    painted_bb = paint_oysters(image_path,bounding_rectangles)
    
    if plot:
        plt.figure(figsize=(20,10))
        plt.imshow(img)
        plt.title("original")
        
        plt.figure(figsize=(20,10))
        plt.imshow(bilateral_filtered,cmap="gray")
        plt.title("bilateral filtered, grayscaled")
        
        plt.figure(figsize=(20,10))
        plt.imshow(thresh,cmap="gray")
        plt.title("threshold")
        
        plt.figure(figsize=(20,10))
        plt.imshow(edges,cmap="gray")
        plt.title("edges")
        
        plt.figure(figsize=(20,10))
        plt.imshow(gaussian_blurred,cmap="gray")
        plt.title("gaussian blurred")
        
        plt.figure(figsize=(20,10))
        plt.imshow(dilated,cmap="gray")
        plt.title("dilated")
        
        plt.figure(figsize=(20,10))
        img_gray = cv2.drawContours(img_gray, contours, -1, (0,255,0), 12)
        plt.imshow(img_gray,cmap="gray")
        plt.title("Contours")
        
        plt.figure(figsize=(20,10))
        plt.imshow(painted_bb)
        plt.title("Bounding boxes")
    
    return [Image.fromarray(img),Image.fromarray(bilateral_filtered),
            Image.fromarray(thresh),Image.fromarray(edges_out),Image.fromarray(gaussian_blurred),Image.fromarray(dilated),
            Image.fromarray(cv2.drawContours(img_gray, contours, -1, (0,255,0), 20)), Image.fromarray(painted_bb)]
    
def predict_and_paint_oysters(image_path, bounding_rectangles, model):
    
    image = plt.imread(image_path)
    
    for (x,y,w,h) in bounding_rectangles:
        #if ((h/w > 0.75) and (h/w < 1.5)):
        if (h > 200) and (w > 200):
            pred = predict_oyster_image(image[y:y+h,x:x+w,:],model)
            if pred > 0.5:
                cv2.rectangle(image, (x, y), (x+w, y+h), (0,255,0), 10)
                cv2.putText(image, str(np.round(pred[0],2)), (x+4,y+50), cv2.FONT_HERSHEY_SIMPLEX,  
                   2.5, (0,255,0), 3, cv2.LINE_AA) 

            elif pred <= 0.5:
                cv2.rectangle(image, (x, y), (x+w, y+h), (255,0,0), 10)
                cv2.putText(image, str(np.round(pred[0],2)), (x+4,y+50), cv2.FONT_HERSHEY_SIMPLEX,  
                   2.5, (255,0,0), 3, cv2.LINE_AA)
    return image
    
    
def paint_oysters(image_path,bounding_rectangles):
    image = plt.imread(image_path)
    
    for (x,y,w,h) in bounding_rectangles:
        #if ((h/w > 0.75) and (h/w < 1.5)):
        if (h > 200) and (w > 200):
            cv2.rectangle(image, (x, y), (x+w, y+h), (0,255,0), 10)

    return image

    
def predict_oyster_image(oyster_image, model):
    (_,h,w,_) = model.input_shape
    img = Image.fromarray(oyster_image).resize((w,h))
    img = np.expand_dims(img, axis=0)/255
    pred = model.predict(img)
    return pred.flatten()
    
    
def adjust_gamma(image, gamma=1.0):
	# build a lookup table mapping the pixel values [0, 255] to
	# their adjusted gamma values
	invGamma = 1.0 / gamma
	table = np.array([((i / 255.0) ** invGamma) * 255
		for i in np.arange(0, 256)]).astype("uint8")
 
	# apply gamma correction using the lookup table
	return cv2.LUT(image, table)


def crop_and_save_oyster_image(filepath,contours,storepath):
    temp_image = cv2.imread(filepath)
    species = get_image_number(filepath)
    print(filepath)
    for idx, (x,y,w,h) in enumerate(contours):
        if (h > 200) and (w > 200):
            temp_store_path = os.path.join(storepath, os.path.basename(filepath)[:-4] + "_" + 
                                           str(idx) + "_" + species + "_" + ".jpg")
            cv2.imwrite(temp_store_path, temp_image[y:y+h,x:x+w,:])
            
def get_concat_h(im1, im2):
    dst = Image.new('RGB', (im1.width + im2.width, im1.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (im1.width, 0))
    return dst

def get_concat_v(im1, im2):
    dst = Image.new('RGB', (im1.width, im1.height + im2.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (0, im1.height))
    return dst